package ru.dbotthepony.mc.otm.capability.energy

import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.NumericTag
import net.minecraft.nbt.Tag
import net.minecraftforge.common.util.INBTSerializable
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.core.forValidRefs
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.math.set
import ru.dbotthepony.mc.otm.core.nbt.map
import ru.dbotthepony.mc.otm.core.nbt.set
import java.lang.ref.WeakReference
import java.util.*
import kotlin.collections.ArrayList

class ProfiledEnergyStorage<E : IMatteryEnergyStorage>(val parent: E) : IMatteryEnergyStorage, INBTSerializable<CompoundTag?> {
	init {
		storages.add(WeakReference(this))
	}

	override fun extractEnergy(howMuch: Decimal, simulate: Boolean): Decimal {
		return recordTransfer(parent.extractEnergy(howMuch, simulate), simulate)
	}

	override fun receiveEnergy(howMuch: Decimal, simulate: Boolean): Decimal {
		return recordReceive(parent.receiveEnergy(howMuch, simulate), simulate)
	}

	override fun extractEnergyChecked(howMuch: Decimal, simulate: Boolean): Decimal {
		return recordTransfer(parent.extractEnergyChecked(howMuch, simulate), simulate)
	}

	override fun receiveEnergyChecked(howMuch: Decimal, simulate: Boolean): Decimal {
		return recordReceive(parent.receiveEnergyChecked(howMuch, simulate), simulate)
	}

	override val canSetBatteryLevel: Boolean get() = parent.canSetBatteryLevel
	override val missingPower: Decimal get() = parent.missingPower

	override fun drainBattery(): Boolean {
		return parent.drainBattery()
	}

	override fun fillBattery(): Boolean {
		return parent.fillBattery()
	}

	override var batteryLevel: Decimal
		get() = parent.batteryLevel
		set(value) { parent.batteryLevel = value }

	override val maxBatteryLevel: Decimal get() = parent.maxBatteryLevel
	override val energyFlow: FlowDirection get() = parent.energyFlow

	private val historyReceiveInternal = ArrayList<Decimal>(HISTORY_SIZE)
	private val historyTransferInternal = ArrayList<Decimal>(HISTORY_SIZE)

	val historyReceive: List<Decimal> = Collections.unmodifiableList(historyReceiveInternal)
	val historyTransfer: List<Decimal> = Collections.unmodifiableList(historyTransferInternal)

	val lastTickReceive
		get() = historyReceiveInternal[tick]

	val lastTickTransfer
		get() = historyTransferInternal[tick]

	private var thisTickReceive = Decimal.ZERO
	private var thisTickTransfer = Decimal.ZERO

	var tick = 0
		private set

	init {
		for (i in 0 until HISTORY_SIZE) {
			historyReceiveInternal.add(Decimal.ZERO)
			historyTransferInternal.add(Decimal.ZERO)
		}
	}

	private fun recordTransfer(value: Decimal, simulate: Boolean): Decimal {
		if (!simulate) {
			thisTickTransfer += value
		}

		return value
	}

	private fun recordReceive(value: Decimal, simulate: Boolean): Decimal {
		if (!simulate) {
			thisTickReceive += value
		}

		return value
	}

	private fun tick() {
		tick = (tick + 1) % HISTORY_SIZE

		historyReceiveInternal[tick] = thisTickReceive
		historyTransferInternal[tick] = thisTickTransfer

		thisTickReceive = Decimal.ZERO
		thisTickTransfer = Decimal.ZERO
	}

	val savedata: INBTSerializable<CompoundTag?> = object : INBTSerializable<CompoundTag?> {
		override fun serializeNBT(): CompoundTag {
			return CompoundTag().also {  tag ->
				tag["historyReceive"] = ListTag().also {
					for (value in historyReceiveInternal) {
						it.add(value.serializeNBT())
					}
				}

				tag["historyTransfer"] = ListTag().also {
					for (value in historyTransferInternal) {
						it.add(value.serializeNBT())
					}
				}

				tag["historyTick"] = tick
			}
		}

		override fun deserializeNBT(nbt: CompoundTag?) {
			tick = 0
			thisTickReceive = Decimal.ZERO
			thisTickTransfer = Decimal.ZERO

			for (i in 0 until HISTORY_SIZE) {
				historyReceiveInternal[i] = Decimal.ZERO
				historyTransferInternal[i] = Decimal.ZERO
			}

			nbt ?: return

			nbt.map("historyTick") { it: NumericTag ->
				tick = it.asInt
			}

			nbt.map("historyReceive") { it: ListTag ->
				for (i in 0 until HISTORY_SIZE.coerceAtMost(it.size)) {
					historyReceiveInternal[i] = Decimal.deserializeNBT(it[i])
				}
			}

			nbt.map("historyTransfer") { it: ListTag ->
				for (i in 0 until HISTORY_SIZE.coerceAtMost(it.size)) {
					historyTransferInternal[i] = Decimal.deserializeNBT(it[i])
				}
			}
		}
	}

	override fun serializeNBT(): CompoundTag {
		val tag: CompoundTag

		if (parent is INBTSerializable<*>) {
			tag = (parent as INBTSerializable<CompoundTag?>).serializeNBT() ?: CompoundTag()
		} else {
			tag = CompoundTag()
		}

		val tag2 = savedata.serializeNBT()!!

		for (k in tag2.allKeys) {
			tag[k] = tag2[k]!!
		}

		return tag
	}

	override fun deserializeNBT(nbt: CompoundTag?) {
		if (parent is INBTSerializable<*>) {
			(parent as INBTSerializable<CompoundTag?>).deserializeNBT(nbt)
		}

		savedata.deserializeNBT(nbt)
	}

	companion object {
		const val HISTORY_SIZE = 20
		private val storages = ArrayList<WeakReference<ProfiledEnergyStorage<*>>>()

		// после раздумий, наиболее корректное место для обновления состояния профилированных
		// хранилищ энергии будет после всех остальных хуков на тики после тика сервера
		// разумеется, такое решение может несколько снизить производительность сервера,
		// но ничего страшного произойти не должно
		internal fun onServerPostTick() {
			storages.forValidRefs { it.tick() }
		}
	}
}
